class AeneasError(Exception):
    """ Aeneas base exception """

    def __init__(self, msg: str):
        self.msg = msg

    def __str__(self):
        return self.msg


class AeneasResponseError(AeneasError):
    def __init__(self, text: str, sql_code: str):
        self.text = text
        self.sql_code = sql_code

    def __str__(self):
        return str(self.__dict__)


class AeneasConnectionError(AeneasError):
    pass


class AeneasActionError(AeneasError):
    pass



import asyncio
from aiohttp import client_exceptions, ClientSession, ClientTimeout
from . import exceptions

TIMEOUT = 10


class Aeneas(object):
    """ Aeneas fullnode API client """

    def __init__(self, uri: str = 'ws://localhost:9085/aeneas'):
        """
        Client initialization

        :param uri: WebSocket server uri
        """

        self._socket = None
        self._session = None
        self._uri = uri
        self._response = {}
        self._timeout = ClientTimeout(total=TIMEOUT)

        # create an event loop and WebSocket connection

        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._loop.run_until_complete(self._create_socket())

    def __del__(self):
        self._loop.run_until_complete(self._close_socket())
        self._loop.close()

    async def _create_socket(self):
        """ Establish a WebSocket connection """

        self._session = ClientSession(timeout=self._timeout)

        try:
            self._socket = await self._session.ws_connect(self._uri)

        except client_exceptions.ClientConnectorError:
            raise exceptions.AeneasConnectionError('Connection to given uri failed, uri: {}'.format(self._uri))

        except client_exceptions.ServerConnectionError as e:
            raise exceptions.AeneasConnectionError('Server connection error: {}'.format(e))

    async def _close_socket(self):
        """ Close a WebSocket connection """

        if self._socket is not None:
            await self._socket.close()

        await self._session.close()

    async def _request(self, data: dict, reestablish: bool = False) -> None:
        """
        Send request via WebSocket connection

        :param data: request data dictionary
        :param reestablish: indicate attempt to reestablish connection
        """

        try:
            await self._socket.send_json(data)
            self._response = await self._socket.receive_json()

        except client_exceptions.ServerConnectionError as e:
            raise exceptions.AeneasConnectionError('Server connection error: {}'.format(e))

        except TypeError:
            if reestablish:
                msg = 'The server closed the connection. Attempt to re-establish connection failed.'
                raise exceptions.AeneasConnectionError(msg)

            await self._close_socket()
            await self._create_socket()
            await self._request(data, reestablish=True)

    def _parse_errors(self):
        """ Parse Aeneas fullnode API response errors """

        if not len(self._response):
            return

        if self._response['action'] == 'ErrorResponse':
            raise exceptions.AeneasActionError('Wrong action, message: {}'.format(self._response['msg']))

        if self._response.get('status') == 'error':
            data = self._response['exception']
            raise exceptions.AeneasResponseError(data['text'], data['sqlCode'])

    def execute(self, action: str, params: dict = None) -> dict:
        """
        Make request with event loop

        :param action: Aeneas fullnode API action name
        :param params: dictionary with request parameters
        :return: json response
        """

        data = {'msg': {'action': action}}

        if params is not None:
            data['msg'].update(params)

        self._loop.run_until_complete(self._request(data))
        self._parse_errors()
        return self._response



from setuptools import setup

with open('README.rst') as file:
    long_description = file.read()

setup(
    name='pyaeneas',
    packages=['pyaeneas'],
    version='0.2',
    license='MIT',
    description='Aeneas fullnode API Python client',
    long_description=long_description,
    author='Andrii Matiiash',
    author_email='andmati743@gmail.com',
    url='https://github.com/AverHLV/pyaeneas',
    download_url='https://github.com/AverHLV/pyaeneas/archive/0.2.tar.gz',
    keywords=['ASYNC', 'AENEAS', 'CLIENT', 'BLOCKCHAIN'],

    install_requires=[
        'aiohttp',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)
